#ifndef READCONFIG_C
#define READCONFIG_C

#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <ctype.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <errno.h>
#include <arpa/tftp.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "types.h"
#include "config.h"
#define	 GLOBDEF
#include "globals.h"
#undef   GLOBDEF
#include "util.h"
#include "dll.h"
#include "dllapps.h"
#include "access.h"
#include "logfun.h"
#include "getarg.h"
#include "readconfig.h"

#define CMD_UNKNOWN		0
#define CMD_DEFAULT_DIR		1
#define CMD_MAX_SYSLOG		2
#define CMD_MESSAGE_LEVEL	3
#define CMD_ROOT_DIR		4
#define CMD_ACCESS_LIST		5
#define CMD_RESTRICT		7
#define CMD_INPUT_WAIT		8
#define CMD_USER		9
#define CMD_GROUP		10
#define CMD_THREAD		11
#define CMD_FILTER		12
#define CMD_MAP			13

typedef struct CMDS_s { CHAR *Name;
                        INT Code;
                        INT minArgs;
			INT intCheck; } CMDS_t;

static CMDS_t configCmds[] = { 
		{ "defaultdirectory", 	CMD_DEFAULT_DIR,	2,	0 } ,
		{ "maxsyslog",	 	CMD_MAX_SYSLOG, 	2,	1 } ,
		{ "messagelevel", 	CMD_MESSAGE_LEVEL, 	2,	1 } ,
		{ "rootdirectory", 	CMD_ROOT_DIR, 		2,	0 } ,
		{ "accesslist", 	CMD_ACCESS_LIST, 	4,	1 } ,
		{ "restrict", 		CMD_RESTRICT, 		3,	1 } ,
		{ "idlewait", 		CMD_INPUT_WAIT, 	2,	1 } ,
		{ "user", 		CMD_USER, 		2,	0 } ,
		{ "group", 		CMD_GROUP, 		2,	0 } ,
		{ "limit", 		CMD_THREAD, 		2,	1 } ,
		{ "filter", 		CMD_FILTER, 		3,	1 } ,
		{ "map", 		CMD_MAP, 		4,	1 } ,
		{ NULL, 		CMD_UNKNOWN, 		0,	0 } };

static INT curLine;
static INT threads = 0;
static FILE *configFILE, *dumpFILE;
static struct stat configstat;

/*
	print copyright info
*/

static VOID copyright(FILE *fp, INT short_form)
{
#define COPYRIGHT1 \
"Copyright (C) 1998,1999 by A.G.Schindler, Darmstadt, Germany (schindler@az1.de)\n\n" 
#define COPYRIGHT2 \
"ACL-tftpd is free software; you can redistribute it and/or modify it under\n" \
"the terms of the GNU General Public License as published by the Free\n" \
"Software Foundation; either version 2, or (at your option) any later version.\n\n" \
"ACL-tftpd is distributed in the hope that it will be useful,\n" \
"but WITHOUT ANY WARRANTY; without even the implied warranty of\n" \
"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n" \
"See the GNU General Public License for more details.\n\n" \
"You should have received a copy of the GNU General Public License along\n" \
"with ACL-tftpd -- see the file COPYING.\n" \
"If not, write to the Free Software Foundation\n" \
"59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.\n\n"

    fprintf(fp, COPYRIGHT1);
    if (!short_form)
        fprintf(fp, COPYRIGHT2);
}

/*
	check if a file is identical to our config file
*/

INT isConfigFile (struct stat *sb)
{
    return (sb->st_dev == configstat.st_dev && 
            sb->st_ino == configstat.st_ino);
}

/*
	force parsing of config file at start of next transaction
*/

VOID refreshConfig (VOID)
{
    memset (&configstat, 0, sizeof(configstat));
}

/*
	get the file descriptor of the config file
*/

INT getConfigFd (VOID)
{
    if (configFILE == NULL)
        return -1;

    return (fileno(configFILE));
}



/*
	extended string compare
	- case insensitive
	- ignores embedded dashes and underscores
*/

static INT cmpXtended (CHAR *s1, CHAR *s2)
{
    CHAR c1, c2;

    for (; c1 = *s1, c2 = *s2; s1++, s2++) {
        while (c1 == '-' || c1 == '_')
    	    c1 = *(++s1);

        if (isupper (c1))
	    c1 = tolower (c1);

        while (c2 == '-' || c2 == '_')
	    c2 = *(++s2);

        if (isupper (c2))
	    c2 = tolower (c2);

        if (c1 != c2 || c1 == '\0')
	    break;
    }
    return (INT) (c1 - c2);
}

/*
	read a line from the config file
	deal with continuation constructs and
	delete comments
*/

static CHAR *getConfigLine(FILE *fd)
{
#define INTSIZ 4096
    static CHAR line[INTSIZ], *ptr, *start;
    INT rest, len;

    for (ptr = line, rest = INTSIZ; fgets(ptr, rest, fd) != NULL;) {
        curLine++;
        for (start = ptr; *start == ' ' || *start == '\t'; start++);
        if (*start == '#') continue;
        len = strlen(ptr) - 1;
        if (len < 1) continue;
        ptr += len;
        if (*(ptr-1) != '\\') break;
        len--;
        ptr--;
        rest -= len;
    }
    *ptr = '\0';
    return ptr == line ? NULL : line;
}

/*
	scan the command list for the current token
*/

static CMDS_t *findCmd (CHAR *token)
{
    CMDS_t *cp;

    for (cp = configCmds; cp->Name != NULL; cp++)
        if (!cmpXtended (cp->Name, token))
  	    break;

    return cp;
}


/*
	config error prefix
*/

static CHAR *configErr(VOID)
{
    static CHAR confErrString[64];

    if (curLine)
        sprintf(confErrString, "Config error in line %d:", curLine);
    else
        strcpy(confErrString, "Config error:");

    return confErrString;
}

/*
	allocate space for a string and copy it there, and return length 
*/

static INT newStringLen (CHAR *arg, CHAR **new, INT *length)
{
    CHAR *tmp;

    if (arg == NULL) {
        *new = NULL;
        *length = 0;
    }
    else {
        if ((tmp = strdup(arg)) != NULL) {
            *length = strlen(tmp);
            *new    = tmp;
            return 0;
        }
    }

    return 1;
}

INT cutConfigLine (CHAR *line, INT *argc, CHAR **argv, INT maxargs, CHAR **ctx)
{
#define SEP	" \t"
#define	QUOTE	"\"\"''"

    static CHAR *tok;
    INT i=0;

    for (tok=qstrtok(line,SEP,ctx,QUOTE); tok != NULL; 
         tok=qstrtok(NULL,SEP,ctx,QUOTE)) {
        if (i >= maxargs) {
            LOG_MSG1 (LOG_ERR, "%s too many arguments (> %d)", 
                      configErr(), maxargs);
            return -1;
        }
        argv[i++] = tok;
    }
    *argc = i;
    return i;

#undef SEP
}

/*
	open the dump file
*/

static VOID openDump()
{
    if (spMode || CheckOnly)
        dumpFILE = stderr;
    else
        dumpFILE = fopen (DumpFile,"w+");
}
           
/*
	close the dump file
*/

static VOID closeDump()
{
    if (dumpFILE != stderr && dumpFILE != NULL)
        fclose (dumpFILE);
}

/*
	dump the current configuration parameters
*/

static VOID configDump(VOID)
{
    struct passwd *pp;
    struct group  *gp;

    pp = getpwuid(childUid);
    gp = getgrgid(childGid);

    fprintf(dumpFILE, "\n"
                "Message Level    : %d\n"
                "Max SYSLOG Level : %d\n"
                "Message Override : %s\n"
                "Idle Timeout     : %d min.\n"
                "Timeout Override : %s\n"
                "User             : %-5d %s\n"
                "Group            : %-5d %s\n"
                "Default Dir.     : %s\n"
                "Root Directory   : %s\n",
                MsgLevel, MaxSyslog,
                MsgOverride < 0 ? "No" : "Yes",
                MaxInputWait/60,
                tftpWaitOverride < 0 ? "No" : "Yes",
                childUid, pp == NULL ? " * unknown *": pp->pw_name,
                childGid, gp == NULL ? " * unknown *": gp->gr_name,
                DefaultDirectory == NULL ? "-" : DefaultDirectory,
                RootDirectory    == NULL ? "-" : RootDirectory);
}

/*
	print version info
*/

static VOID version(FILE *fp)
{
    fprintf(fp, "DARPA tftp protocol server '%s' V-%d.%2.2d%s%s\n",
                PACKAGE_NAME, MAJOR_VERSION, MINOR_VERSION,
                strlen(PATCH_LEVEL) ? " patch level " : "",
                strlen(PATCH_LEVEL) ? PATCH_LEVEL : "");
}

/*
	dump info on open file descriptors
*/

VOID fdDump(VOID)
{
    INT fd, cnt, soc, coe;
    struct stat statBuf;

    for (fd = 0, cnt = 0; fd < 32; fd++) {
        if (fstat(fd, &statBuf)) 
            continue;
        if (!cnt++) {
            fprintf (dumpFILE, "\nOpen file descriptors:\n"
                               "Descr.\tSocket\tClose-on-exec\n");
        }
        soc = isaSocketFd (fd);
        coe = fcntl (fd, F_GETFD);
        if (coe < 0)
            coe = 0;
        fprintf (dumpFILE,"%d\t%s\t%s\n",
                 fd, soc ? "yes" : "no", coe ? "yes" : "no" );
    }
}

/*
	dump configuration and access lists
*/

VOID doDump(VOID)
{
    time_t itime;

    time(&itime);
    openDump();
    version(dumpFILE);
    copyright(dumpFILE, 1);
    fprintf(dumpFILE, "Configuration dump generated by PID %d at %s",
            getpid(), asctime(localtime(&itime)));
    configDump();
    fdDump();
    accessDump(dumpFILE);
    filterDump(dumpFILE);
    threadDump(dumpFILE);
    closeDump();
}

/* 
	reset configuration options to defaults
	this function does no changes on the process environment. 
*/

static INT existentDir (CHAR *dirname)
{
    struct stat statb;

    return ( stat(dirname, &statb) < 0 ? 0 : 1 );
}


static INT resetConfig ()
{
    if (restrictListH != NULL)
        deleteReChain (restrictListH);

    if ((restrictListH = initReChain()) == NULL)
        return -1;

    if (accessListH != NULL)
        deleteACLChain (accessListH);

    if ((accessListH = initACLChain()) == NULL)
        return -1;

    if (filterListH != NULL)
        deleteFilterChain (filterListH);

    if ((filterListH = initFilterChain()) == NULL)
        return -1;

    if (mapListH != NULL)
        deleteReChain (mapListH);

    if ((mapListH = initReChain()) == NULL)
        return -1;

    if (threadListH != NULL)
        deleteACLChain (threadListH);

    if ((threadListH = initACLChain()) == NULL)
        return -1;

    threads = 0;

    if (procListH != NULL)
        deletePidChain (procListH);

    if ((procListH = initPidChain()) == NULL)
        return -1;

    if (MsgOverride < 0)			/* reset message level */
        MsgLevel = MESSAGELEVEL;

    if (MslOverride < 0)			/* reset max. syslog level */
        MaxSyslog = MAXSYSLOG;
    
    if (MaxSyslog < 0)				/* same as message level */
        MaxSyslog = MsgLevel;			/* if < 0                */

    if (tftpWaitOverride < 0)			/* reset idle timeout */
        MaxInputWait = DEFTIMEOUT*60;

    if (DefaultDirectory != NULL) 		/* reset default directory */
	  free (DefaultDirectory);

    if (DEFAULTDIR == NULL) {
        DefaultDirectory       = NULL;
        DefaultDirectoryLength = 0;
    }
    else if (existentDir (DEFAULTDIR)) {
        if (newStringLen(DEFAULTDIR, &DefaultDirectory, 
                                     &DefaultDirectoryLength))
            if (!CheckOnly)
                return 1;
    }

    if (RootDirectory != NULL) { 		/* reset root directory */
        free (RootDirectory);
    }

    if ((ROOTDIR == NULL) || (spMode)) {
        RootDirectory       = NULL;
        RootDirectoryLength = 0;
        changeRoot          = 0;
    }
    else if (existentDir (ROOTDIR)) {
        if (newStringLen (ROOTDIR, &RootDirectory, 
                                   &RootDirectoryLength))
            if (!CheckOnly)
                return 1;
        changeRoot = 1;
    }

    childUid = TFTPUID;				/* reset user & group */
    childGid = TFTPGID;

    return 0;
}


/* 
	(re)-read the tftpd config file.  
	We keep an open handle on the file
	and check to see if it's been modified since the last read.
*/

INT readConfigFile (CHAR *ConfigFileName)
{
#define MAXARGS 64

    struct stat statb;
    CHAR *line;
    CMDS_t *cmdp;
    INT perm, list;
    INT cargc;
    CHAR *cargv[MAXARGS];
    CHAR *ctx;
    INT rtc = 0;
    INT fnum;
    INT MLevel, MSyslog;

    if (ConfigFileName == NULL)
        ConfigFileName = CONFIGFILE;


    if (configFILE == NULL) {
        if ((configFILE = fopen (ConfigFileName, "r")) == NULL) {
	    LOG_MSG1 (LOG_ERR, "Cannot open config file '%s': %s",
                      ConfigFileName, strerror (errno));
            return 1;
        }
        if (fstat(fileno(configFILE), &configstat) < 0) {
            LOG_MSG1 (LOG_ERR, "Cannot get config file status: %s",
                      strerror (errno));
            return 1;
        }
    }
/*
	See if the file has been changed since the last time we read it.
*/
    else {

        if (fstat(fileno(configFILE), &statb) < 0) {
            LOG_MSG1 (LOG_ERR, "Cannot get config file status: %s",
                      strerror (errno));
            return 1;
        }

        if (statb.st_mtime == configstat.st_mtime &&
            statb.st_ino   == configstat.st_ino ) {
	    LOG_MSG5 (LOG_DEBUG, "Config file '%s' unchanged", ConfigFileName);
            return 0;
        }

        configstat = statb;
    }

    MLevel  = MsgLevel;
    MSyslog = MaxSyslog;

/* 
	Read the config file 
*/
    LOG_MSG4 (LOG_DEBUG, "Processing config file '%s'",ConfigFileName);

    rewind (configFILE);
    curLine = 0;

    if (resetConfig ()) {
        errNoDynmem(NULL);
        return 1;
    }

    while ((line = getConfigLine(configFILE)) != NULL) {

	if (cutConfigLine (line, &cargc, cargv, MAXARGS, &ctx) < 0)
            continue;

        cmdp = findCmd (cargv[0]);

/*
	check config clause for sufficient number of parameters
*/
        if (cargc < cmdp->minArgs) {
            LOG_MSG1 (LOG_ERR, "%s not enough arguments to '%s' clause",
                      configErr(), cargv[0]);
            rtc++;
            if (CheckOnly)
                continue;
            else
                return 1;
        }

/*
	check numeric argument if necessary
*/
        if (cmdp->intCheck) {
            if (sscanf(cargv[1], "%u", (unsigned *)&list) != 1) {
                LOG_MSG1 (LOG_ERR, "%s illegal numeric argument '%s'"
                          " in '%s' clause",
                          configErr(), cargv[1], cargv[0]);
                rtc++;
                if (CheckOnly)
                    continue;
                else
                    return 1;
            }
        }

        switch (cmdp->Code) {

	    case CMD_DEFAULT_DIR:	/* default directory */

                if (newStringLen (cargv[1], &DefaultDirectory,
                                            &DefaultDirectoryLength)) {
                    errNoDynmem(NULL);
                    return 1;
                }

                if (!existentDir(DefaultDirectory)) {
                    LOG_MSG1 (LOG_ERR, "Default directory '%s' not found",
                              DefaultDirectory);
                    rtc++;
                    if (!CheckOnly)
                        return 1;
                }
                break;


            case CMD_ROOT_DIR:		/* virtual root directory */

                if (spMode) {
                    LOG_MSG1 (LOG_WARNING, "RootDirectory clause invalid"
                              " in Single-process mode -- ignored");
                }
                else {
                    if (newStringLen (cargv[1], &RootDirectory,
                                                &RootDirectoryLength)) {
                        errNoDynmem(NULL);
                        return 1;
                    }

                    if (!existentDir(RootDirectory)) {
                        LOG_MSG1 (LOG_ERR, "Root directory '%s' not found",
                                  RootDirectory);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                    }
                    changeRoot = 1;
                }
                break;


            case CMD_MAX_SYSLOG:	/* set max. syslog message level */

                MSyslog = list;
                break;


            case CMD_MESSAGE_LEVEL:	/* set message level */

                MLevel = list;
                break;


            case CMD_ACCESS_LIST:	/* create or extend an access list */

                if ((perm = evalPermission(cargv[2])) < 0) {
                    LOG_MSG1 (LOG_ERR, "%s illegal permission list '%s'",
                              configErr(), cargv[2]);
                    rtc++;
                    if (!CheckOnly)
                        return 1;
                }

                switch (addACL(accessListH, list, perm, cargv[3], 
                               cargc >= 5 ? cargv[4] : C_IPMASK_HOST, NULL)) {

                    case -1:
                        if (MsgLevel)
                            errNoDynmem(NULL);
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, "%s unresolvable hostname '%s'",
                                  configErr(), cargv[3]);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                        break;
                }
                break;


            case CMD_RESTRICT:	/* restrict file access by access list */

                if (cargc > 3) {
                    if (sscanf(cargv[3], "%u", (unsigned *)&fnum) != 1) {
                        LOG_MSG1 (LOG_ERR, "%s illegal numeric argument '%s'"
                                           " in 'restrict' clause",
                                           configErr(), cargv[3]);
                        rtc++;
                        if (CheckOnly)
                            continue;
                        else
                            return 1;
                    }
                }
                else {
                    fnum = 0;
                }

                switch (addRePattern(restrictListH, list, cargv[2],
                                     NULL, fnum, NULL)) {

                    case -1:
                        if (MsgLevel)
                            errNoDynmem(NULL);
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, "%s illegal regular expression '%s'",
                                  configErr(), cargv[2]);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                        break;
                }

                if (getACLByMark(accessListH, NULL, list) == NULL)
                    LOG_MSG2 (LOG_WARNING,
                              "%s reference to undefined access list '%d'",
                              configErr(), list);

                if (fnum && (getFilterByMark(filterListH, NULL, fnum) == NULL))
                    LOG_MSG2 (LOG_WARNING,
                              "%s reference to undefined filter '%d'",
                              configErr(), fnum);
                break;


            case CMD_INPUT_WAIT:	/* set input timeout period */

                if (tftpWaitOverride < 0)
                    MaxInputWait = list*60;
                break;


            case CMD_USER:		/* set new user */

                {
                    struct passwd *sp;
                    if ((sp = getpwnam (cargv[1])) == NULL) {
                        LOG_MSG1 (LOG_ERR, "unknown tftp user %s", cargv[1]);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                    }
                    else
                        childUid = sp->pw_uid;
                }
                break;

	    case CMD_GROUP:		/* set new group */

                {
                    struct group *sg;
                    if ((sg = getgrnam (cargv[1])) == NULL) {
                        LOG_MSG2 (LOG_ERR, "unknown tftp group %s", cargv[1]);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                    }
                    else
                        childGid = sg->gr_gid;
                }
                break;

            case CMD_THREAD:	/* add a group limit */

                switch (addACL(threadListH, ++threads, list, cargv[2], 
                               cargc >= 4 ? cargv[3] : C_IPMASK_HOST, NULL)) {

                    case -1:
                        if (MsgLevel)
                            errNoDynmem(NULL);
                        return 1;

                    case -2:
                        LOG_MSG1 (LOG_ERR, "%s unresolvable hostname '%s'",
                                  configErr(), cargv[3]);
                        rtc++;
                        if (!CheckOnly)
                            return 1;
                        break;
                }
                break;

            case CMD_FILTER:	/* add a filter description */

                if (list <=0) {
                    LOG_MSG1 (LOG_ERR, "%s illegal filter %d",
                                        configErr(), list);
                        rtc++;
                        if (CheckOnly)
                            continue;
                        else
                            return 1;
                }

                if (getFilterByMark(filterListH, NULL, list) != NULL) {
                    LOG_MSG1 (LOG_ERR, "%s duplicate filter declaration %d",
                                        configErr(), list);
                        rtc++;
                        if (CheckOnly)
                            continue;
                        else
                            return 1;
                }

                if (cargc < 4)
                    cargv[3] = NULL;

                if (addFilter(filterListH,list,cargv[2],cargv[3],NULL)) {
                    if (MsgLevel)
                        errNoDynmem(NULL);
                    return 1;
                }
                break;


            case CMD_MAP:	/* file map filter entry	*/

                if (checkFileMap(cargv[3])) {
                    LOG_MSG1 (LOG_ERR, "%s illegal file map '%s'",
                                        configErr(), cargv[3]);
                        rtc++;
                        if (CheckOnly)
                            continue;
                        else
                            return 1;
                }

                if (addRePattern(mapListH,list,cargv[2],cargv[3],0,NULL)) {
                    if (MsgLevel)
                        errNoDynmem(NULL);
                    return 1;
                }
                break;


            default:		/* illegal configuration options */

                LOG_MSG1 (LOG_ERR, "%s Illegal configuration keyword '%s'",
                          configErr(), cargv[0]);
                rtc++;
                if (!CheckOnly)
                    return 1;
                break;
	}
    }

    curLine = 0;

    if (MsgOverride < 0)
        MsgLevel = MLevel;

    if (MslOverride < 0) {
        if (MSyslog < 0)
            MaxSyslog = MsgLevel;
        else
            MaxSyslog = MSyslog;
    }

/*
	sort the filter declarations in ascending order
*/
    sortFilterByMark(filterListH, DLL_SORTASC);

    if (CheckOnly)
          doDump();

     
    LOG_MSG4 (LOG_DEBUG, "Exiting readconfig: %d errors detected", rtc); 
    return rtc;
}

static VOID usage()
{
    fprintf(stderr, 
            "Usage is:\n"
            "%s [-s] [-n] [-m lev] [-c] [-t time] [-l] [-?]"
            " [-q] [-V] [-C conf] [-D dump] [-L log]\n"
            "\t-s\tRun in single process mode (don't fork)\n"
            "\t-n\tInhibit name to address (BIND) lookup\n"
            "\t-m lev\tSet message level to <lev> (default: %d)\n"
            "\t-c\tJust check the configuration and exit\n"
            "\t-t time\tSet idle timeout to <time> minutes (default: %d)\n"
            "\t-l\tAllow secure messages only\n"
            "\t-x max\tLimit syslog messages to priority <= <max> (default: %d)\n"
            "\t-?\tPrint this text\n"
            "\t-q\tSilent access denials (no error packets)\n"
            "\t-V\tPrint copyright and version info\n"
            "\t-C conf\tConfig file (default: \"%s\")\n"
            "\t-D dump\tDump file (default: \"%s\")\n"
            "\t-L log\tOptional log file (default: no log file)\n",
            myName, MESSAGELEVEL, DEFTIMEOUT, 
            (MAXSYSLOG == -1) ? MESSAGELEVEL : MAXSYSLOG,
            CONFIGFILE, DUMPFILE);
}

/*
	process command line options
*/

VOID getMyName (INT argc, CHAR **argv)
{
    CHAR *p;

    if (argc)
        if ((p = strrchr(argv[0], '/')) == NULL)
            myName = argv[0];
        else
            myName = p+1;
    else
        myName = "tftpd";
}


static void mergeCMDoptions(void) 
{
    if(MsgOverride >= 0)
        MsgLevel = MsgOverride;

    if (CheckOnly) {
        MaxSyslog = 0;
    }
    else {
        if(MslOverride >= 0)
            MaxSyslog = MslOverride;
    }
    if(tftpWaitOverride >= 0)
        MaxInputWait = tftpWaitOverride*60;

    if (CheckOnly) {
        if(MsgLevel < 4)
            MsgLevel  = 4;
        MaxSyslog = 0;
    }

    if (spMode || CheckOnly) {
        if(MsgLevel < 3)
            MsgLevel = 3;
        MaxSyslog = 0;
        SafeLog   = 0;
    }

    if (ConfigFile == NULL)
        ConfigFile = CONFIGFILE;

    if (DumpFile == NULL)
        DumpFile = DUMPFILE;

    if (LogFile == NULL)
        LogFile = LOGFILE;
}


VOID evalCommandLine (INT argc, CHAR **argv, INT merge)
{
    static CHAR *allopts = "lnsqt:m:cC:D:L:x:?V";

    getMyName (argc, argv);

    if (getarg (argc, argv, allopts, '?', "", NULL) != NULL) {
        version(stderr);
        usage();
        exit(0);
    }

    if (getarg (argc, argv, allopts, 'V', "", NULL) != NULL) {
        version(stderr);
        copyright(stderr, 0);
        exit(0);
    }

    spMode            = atoi (getarg (argc, argv, allopts, 's', "1", "0"));
    noResolv          = atoi (getarg (argc, argv, allopts, 'n', "1", "0"));
    MsgOverride       = atoi (getarg (argc, argv, allopts, 'm', "1","-1"));
    tftpWaitOverride  = atoi (getarg (argc, argv, allopts, 't', "5","-1"));
    CheckOnly         = atoi (getarg (argc, argv, allopts, 'c', "1", "0"));
    SafeLog           = atoi (getarg (argc, argv, allopts, 'l', "1", "0"));
    MslOverride       = atoi (getarg (argc, argv, allopts, 'x', "1","-1"));
    SilentDeny        = atoi (getarg (argc, argv, allopts, 'q', "1", "0"));
    ConfigFile        =       getarg (argc, argv, allopts, 'C',NULL,NULL);
    DumpFile          =       getarg (argc, argv, allopts, 'D',NULL,NULL);
    LogFile           =       getarg (argc, argv, allopts, 'L',NULL,NULL);

    if (merge)
        mergeCMDoptions();
}        

#ifdef TEST

INT main (INT argc, CHAR **argv)
{
    INT rtc;

    evalCommandLine(argc, argv, 0);

    fprintf (stderr,"\nCommand line options for '%s':\n\n"
            "Single Process   : %s\n"
            "BIND lookup      : %s\n"
            "Message Level    : %d (%s)\n"
            "Max SYSLOG Level : %d %s\n"
            "Idle Timeout     : %d min. (%s)\n"
            "Check only       : %s\n"
            "Safe Log         : %s\n"
            "Silent Deny      : %s\n"
            "Show Help        : %s\n"
            "Config File      : %s %s\n"
            "Dump File        : %s %s\n"
            "Log File         : %s %s\n",
            myName,
            spMode               ? "Yes" : "No",
            noResolv             ? "No"  : "Yes",
            MsgOverride >= 0     ? MsgOverride : MsgLevel,
            MsgOverride >= 0     ? "CMD option" : "Config file",
            MslOverride >= 0     ? MslOverride : MsgLevel, 
            MslOverride >= 0     ? "" : "(default)",
            tftpWaitOverride>= 0 ? tftpWaitOverride : MaxInputWait/60,
            tftpWaitOverride>= 0 ?  "CMD option" : "Config file",
            CheckOnly            ? "Yes" : "No",
            SafeLog              ? "Yes" : "No",
            SilentDeny           ? "Yes" : "No",
            Help                 ? "yes" : "no",
            ConfigFile == NULL   ? CONFIGFILE : ConfigFile,
            ConfigFile == NULL   ? "(default)" : "",
            DumpFile   == NULL   ? DUMPFILE : DumpFile,
            DumpFile   == NULL   ?  "(default)" : "",
            LogFile    == NULL   ? "-" : LogFile,
            LogFile    == NULL   ? "(no file logging)" : "" );

    spMode   = 1;
    dumpFILE = stderr;
    configDump();
    fprintf(dumpFILE,"\n");

    MsgLevel    = 5;
    MsgOverride = 5;
    MaxSyslog   = 0;
    MslOverride = 0;
    CheckOnly   = 1;

    startLog();
    rtc = readConfigFile (ConfigFile);

    fprintf (stderr, "Status of config file processing %d\n",rtc);

    if (MsgLevel < 4 && !CheckOnly) {
          configDump();
	  accessDump(dumpFILE);
	  threadDump(dumpFILE);
    }

    return 0;
}

#endif /* TEST */

#endif /* READCONFIG_C */
